// SPDX-License-Identifier: MIT
pragma solidity ^0.8.13;

import "@openzeppelin/contracts/token/ERC20/IERC20.sol";
import "@openzeppelin/contracts/security/ReentrancyGuard.sol";
import "@openzeppelin/contracts/access/Ownable.sol";

contract PressTheButton is ReentrancyGuard, Ownable {
    IERC20 public paymentToken;
    IERC20 public usdcToken;

    address public lastPresser;
    address public previousWinner;
    address[] public pressers;

    uint256 public totalPresses;
    uint256 public totalTokensUsed;
    uint256 public totalRewardsWon;
    uint256 public lastPressTime;
    uint256 public pressCost;
    uint256 public gameDuration = 1 hours;

    bool public isActive;

    mapping(address => uint256) public pressCount;
    mapping(address => bool) public hasPressed;

    event ButtonPressed(address indexed presser, uint256 time);
    event PrizeClaimed(address indexed winner, uint256 prizeAmount);
    event FundsAdded(uint256 amount, address indexed depositor);
    event DevFundsClaimed(uint256 amount, address indexed owner);

    constructor(uint256 _pressCost) {
        pressCost = _pressCost;
        isActive = false;
    }

    /* VIEW FUNCTIONS */

    // The winner of the previous/latest round, blank if no winner
    function getPreviousWinner() external view returns (address) {
        return previousWinner;
    }

    // Current rewards in contract
    function getCurrentPrize() external view returns (uint256) {
        return usdcToken.balanceOf(address(this));
    }

    // Time until timer ends
    function timeUntilEnd() external view returns (uint256) {
        uint256 timeElapsed = block.timestamp - lastPressTime;
        return timeElapsed >= gameDuration ? 0 : gameDuration - timeElapsed;
    }

    // Function to see how many times an address has pressed
    function getPressCount(address user) external view returns (uint256) {
        return pressCount[user];
    }

    // Function to see how many times the button has been pressed in total
    function getTotalPresses() external view returns (uint256) {
        return totalPresses;
    }

    // Function to see total tokens that have been used up by calling the pressButton function
    function getTotalTokensUsed() external view returns (uint256) {
        return totalTokensUsed;
    }

    // Function to see the total rewards won (USDC)
    function getTotalRewardsWon() external view returns (uint256) {
        return totalRewardsWon;
    }

    /* CALL FUNCTIONS */

    function setPaymentToken(address _paymentToken) external onlyOwner {
        require(address(paymentToken) == address(0), "PaymentToken is already set");
        paymentToken = IERC20(_paymentToken);
    }

    function setUsdcToken(address _usdcToken) external onlyOwner {
        require(address(usdcToken) == address(0), "USDC Token is already set");
        usdcToken = IERC20(_usdcToken);
    }

    function addFundsToPrizePool(uint256 amount) external onlyOwner {
        require(amount > 0, "Amount should be greater than 0");
        require(
            usdcToken.transferFrom(msg.sender, address(this), amount),
            "Transfer failed"
        );
        emit FundsAdded(amount, msg.sender);
    }

    function setPressCost(uint256 _pressCost) external onlyOwner {
        pressCost = _pressCost;
    }

    function pressButton() external nonReentrant {
        require(isActive, "The game is not active");
        require(msg.sender == tx.origin, "Sender cannot be a contract");
        uint256 balance = paymentToken.balanceOf(msg.sender);
        require(balance >= pressCost, "Not enough $TOKEN tokens");
        uint256 allowance = paymentToken.allowance(msg.sender, address(this));
        require(
            allowance >= pressCost,
            "Contract not approved to spend enough $TOKEN tokens"
        );

        uint256 burnAmount = (pressCost * 90) / 100; // 90% of the fee is burnt

        // Transfer the entire pressCost to the contract
        require(
            paymentToken.transferFrom(msg.sender, address(this), pressCost),
            "Transfer of $TOKEN tokens to contract failed"
        );

        // Burn 90% of the pressCost
        require(
            paymentToken.transfer(
                0x000000000000000000000000000000000000dEaD,
                burnAmount
            ),
            "Transfer of $TOKEN tokens to dead address failed"
        );

        if (!hasPressed[msg.sender]) {
            pressers.push(msg.sender);
        }
        hasPressed[msg.sender] = true;
        lastPresser = msg.sender;
        lastPressTime = block.timestamp;

        // Update the press count for the user and the total presses
        pressCount[msg.sender]++;
        totalPresses++;
        totalTokensUsed += pressCost;

        emit ButtonPressed(msg.sender, block.timestamp);
    }

    function claimPrize() external nonReentrant {
        require(isActive, "The game is not active");
        require(
            block.timestamp > lastPressTime + gameDuration,
            "Game is still ongoing"
        );
        require(
            msg.sender == lastPresser,
            "Only the last presser can claim the prize"
        );

        uint256 prizeAmount = usdcToken.balanceOf(address(this));
        require(
            usdcToken.transfer(msg.sender, prizeAmount),
            "Transfer of USDC to winner failed"
        );

        previousWinner = lastPresser; // Update the previous winner
        totalRewardsWon += prizeAmount;

        emit PrizeClaimed(lastPresser, prizeAmount);

        // Reset the game state
        for (uint256 i = 0; i < pressers.length; i++) {
            hasPressed[pressers[i]] = false;
        }
        delete pressers;
        lastPressTime = 0;
        lastPresser = address(0);

        // Set the game to inactive
        isActive = false;
    }

    function setActive(bool _isActive) external onlyOwner {
        isActive = _isActive;
    }

    function setGameDuration(uint256 _duration) external onlyOwner {
        gameDuration = _duration;
    }

    function emergencyResetGame() external onlyOwner {
        require(!isActive, "Game must be inactive to reset");
        for (uint256 i = 0; i < pressers.length; i++) {
            hasPressed[pressers[i]] = false;
        }
        delete pressers;
        lastPressTime = 0;
        lastPresser = address(0);
    }

    function emergencyWithdrawAllTokens(IERC20 token) external onlyOwner {
        uint256 balance = token.balanceOf(address(this));
        require(token.transfer(owner(), balance), "Token transfer failed");
    }

    function claimDevFunds() external onlyOwner {
        uint256 balance = paymentToken.balanceOf(address(this));
        require(
            paymentToken.transfer(owner(), balance),
            "Transfer of $TOKEN tokens to owner failed"
        );

        emit DevFundsClaimed(balance, owner());
    }
}
